"""Guesslang machine learning model"""

import json
import logging
from pathlib import Path
from statistics import mean, stdev
from tempfile import TemporaryDirectory
from typing import List, Tuple, Optional

from guesslang import model


LOGGER = logging.getLogger(__name__)

DATA_DIR = Path(__file__).absolute().parent.joinpath('data')
DEFAULT_MODEL_DIR = DATA_DIR.joinpath('model')
LANGUAGES_FILE = DATA_DIR.joinpath('languages.json')
TEST_REPORT_FILE = 'test-report.json'


class Guess:
    """Guess the programming language of a source code.

    :param model_dir: Guesslang machine learning model directory.
    """

    def __init__(self, model_dir: Optional[str] = None) -> None:
        if model_dir:
            self._saved_model_dir = model_dir
        else:
            self._saved_model_dir = str(DEFAULT_MODEL_DIR)

        try:
            self._model = model.load(self._saved_model_dir)
        except OSError:
            self._model = None

        language_json = LANGUAGES_FILE.read_text()
        language_info = json.loads(language_json)
        self._language_map = {
            name: exts[0] for name, exts in language_info.items()
        }
        self._extension_map = {
            ext: name for name, ext in self._language_map.items()
        }

    @property
    def is_trained(self) -> bool:
        """Check if the current machine learning model is trained.
        Only trained models can be used for prediction.

        :return: the model training status.
        """
        return self._model is not None

    @property
    def supported_languages(self) -> List[str]:
        """List supported programming languages

        :return: language name list.
        """
        return list(self._language_map)

    def language_name(self, source_code: str) -> Optional[str]:
        """Predict the programming language name of the given source code.

        :param source_code: source code.
        :return: the language name
            or ``None`` if no programming language is detected.
        """
        if not source_code.strip():
            LOGGER.warning('Empty source code provided')
            return None

        language_probabilities = self.probabilities(source_code)
        probabilities = [value for _, value in language_probabilities]
        if not self._is_reliable(probabilities):
            LOGGER.warning('No programming language detected')
            return None

        language_name, _ = language_probabilities[0]
        return language_name

    def probabilities(self, source_code: str) -> List[Tuple[str, float]]:
        """Gives the probability that the source code is written
        in each of the supported languages.

        The probabilities are sorted from the most to the least probable
        programming language.

        :param source_code: source code.
        :return: list of language names associated with their probability.
        """
        if not self.is_trained:
            LOGGER.error('Cannot predict using an untrained model')
            raise GuesslangError(
                f'Cannot predict using the untrained model located at '
                f'{self._saved_model_dir}. '
                f'Train your model with `guess.train(source_files_dir)`'
            )

        return model.predict(self._model, self._extension_map, source_code)

    def train(self, source_files_dir: str, max_steps: int) -> float:
        """Train guesslang to recognize programming languages.

        The machine learning model is trained from source code files.
        The files should be split in three subdirectories named
        "train", "valid" and "test".

        :raise GuesslangError: when the training cannot be run.
        :param source_files_dir: directory that contains
            the "train", "valid" and "test" datasets.
        :return: training accuracy, a value between 0 and 1.
        """

        LOGGER.debug('Run safety checks before starting the training')

        if self.is_trained:
            LOGGER.error('Model already trained')
            raise GuesslangError(
                f'The current model located at {self._saved_model_dir} '
                f'is already trained'
            )

        input_path = Path(source_files_dir)
        for dirname in model.DATASET.values():
            dataset_path = input_path.joinpath(dirname)
            if not dataset_path.is_dir():
                LOGGER.error(f'Dataset directory missing {dataset_path}')
                raise GuesslangError(f'No dataset directory: {dataset_path}')

        LOGGER.debug('Run the training')
        extensions = list(self._extension_map)
        with TemporaryDirectory() as model_logs_dir:
            estimator = model.build(model_logs_dir, extensions)
            metrics = model.train(estimator, source_files_dir, max_steps)
            LOGGER.info(f'Training metrics: {metrics}')
            model.save(estimator, self._saved_model_dir)

        LOGGER.debug(f'Test newly trained model {self._saved_model_dir}')
        self._model = model.load(self._saved_model_dir)
        matches = model.test(
            self._model, source_files_dir, self._extension_map
        )

        report_file = Path(self._saved_model_dir).joinpath(TEST_REPORT_FILE)
        json_data = json.dumps(matches, indent=2, sort_keys=True)
        report_file.write_text(json_data)
        LOGGER.debug(f'Test report stored in {report_file}')

        languages = self._language_map.keys()
        total = sum(sum(values.values()) for values in matches.values())
        success = sum(matches[language][language] for language in languages)
        accuracy = success / total
        LOGGER.debug(f'Accuracy = {success} / {total} = {accuracy:.2%}')
        return accuracy

    @staticmethod
    def _is_reliable(probabilities: List[float]) -> bool:
        """Arbitrary rule to determine if the prediction is reliable:

        The predicted language probability must be higher than
        2 standard deviations from the mean.
        """
        threshold = mean(probabilities) + 2*stdev(probabilities)
        predicted_language_probability = max(probabilities)
        return predicted_language_probability > threshold


class GuesslangError(Exception):
    """Guesslang exception class"""



"""Guess the programming language of a given source code"""

from argparse import ArgumentParser, FileType
from copy import deepcopy
import logging.config
import os
import sys
from typing import Any, TextIO, Dict

import tensorflow as tf

from guesslang.guess import Guess, GuesslangError
from guesslang.model import DATASET


LOGGER = logging.getLogger(__name__)
LOGGING_CONFIG = {
    'version': 1,
    'formatters': {
        'simple': {
            'format': '{asctime} {name} {levelname} {message}',
            'datefmt': '%H:%M:%S',
            'style': '{'
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'level': logging.DEBUG,
            'formatter': 'simple',
            'stream': 'ext://sys.stdout'
        }
    },
    'loggers': {
        'guesslang': {
            'level': logging.DEBUG,
            'handlers': ['console'],
            'propagate': 0
        }
    },
    'root': {
        'level': logging.DEBUG,
        'handlers': ['console'],
        'propagate': 0
    },
    'disable_existing_loggers': False
}


def main() -> None:
    """Run command line arguments"""

    # Handle command line arguments
    parser = _build_argument_parser()
    args = parser.parse_args()
    if args.train and (not args.model or not args.steps):
        parser.error('--model and --steps are required when using --train')

    # Setup loggers
    if args.debug:
        logging_level = logging.DEBUG
        tf_logging_level = logging.INFO
        tf_backend_logging_level = '0'
    else:
        logging_level = logging.INFO
        tf_logging_level = logging.ERROR
        tf_backend_logging_level = '3'
    logging_config = _update_config(LOGGING_CONFIG, logging_level)
    logging.config.dictConfig(logging_config)
    tf.get_logger().setLevel(tf_logging_level)
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = tf_backend_logging_level

    # Load the programming language guessing model
    LOGGER.debug(f'Arguments: {args}')
    guess = Guess(args.model)

    try:
        if args.list_supported:
            # List the supported programming languages
            languages = ', '.join(guess.supported_languages)
            print(f'Supported programming languages: {languages}')

        elif args.train:
            # Train from source code files
            LOGGER.debug(f'Train model and save result to: {args.model}')
            accuracy = guess.train(args.train, max_steps=args.steps)
            print(f'Trained model accuracy is {accuracy:.2%}')

        else:
            # Guess source code language
            LOGGER.debug(f'Guess the source code of: {args.filename}')
            content = _read_file(args.filename)
            if args.probabilities:
                # List all the detection probabilities
                scores = guess.probabilities(content)
                texts = (f' {name:20} {score:6.2%}' for name, score in scores)
                table = '\n'.join(texts)
                print(f'Language name       Probability\n{table}')
            else:
                # Print the source code programming language name
                # if it is successfully detected
                language_name = guess.language_name(content)
                if not language_name:
                    language_name = 'Unknown'
                print(f'Programming language: {language_name}')

        LOGGER.debug('Exit OK')

    except GuesslangError:
        LOGGER.critical('Failed!')
        sys.exit(1)

    except KeyboardInterrupt:
        LOGGER.critical('Cancelled!')
        sys.exit(2)


def _build_argument_parser() -> ArgumentParser:
    parser = ArgumentParser(description=__doc__)
    parser.add_argument(
        'filename',
        type=FileType('r'),
        default=sys.stdin,
        nargs='?',
        help="""
            source code file.
            Reads from the standard input (stdin) if no file is given
        """,
    )
    parser.add_argument(
        '-p',
        '--probabilities',
        action='store_true',
        default=False,
        help="""
            show match probability of a given source code
            with each of the supported programming languages
        """,
    )
    parser.add_argument(
        '-l',
        '--list-supported',
        action='store_true',
        default=False,
        help='list the supported programming languages',
    )
    parser.add_argument(
        '--train',
        metavar='TRAINING_DIRECTORY',
        help=f"""
            train from a directory containing source code files.
            The source files should be split in 3 directories named:
            {', '.join(DATASET.values())}.

            --model and --steps values should be provided when using --train
        """
    )
    parser.add_argument(
        '--steps',
        metavar='TRAINING_STEPS',
        type=int,
        help="""
            number of steps training steps. The model accuracy
            and the training time increase with the number of steps
        """,
    )
    parser.add_argument(
        '--model',
        metavar='MODEL_DIR',
        help='custom Guesslang trained model directory',
    )
    parser.add_argument(
        '-d',
        '--debug',
        default=False,
        action='store_true',
        help='show debug messages',
    )
    return parser


def _update_config(config: Dict[str, Any], level: int) -> Dict[str, Any]:
    logging_config = deepcopy(config)
    logging_config['root']['level'] = level
    logging_config['loggers']['guesslang']['level'] = level
    return logging_config


def _read_file(input_file: TextIO) -> str:
    if input_file is sys.stdin:
        LOGGER.debug('Write your source code here. End with CTR^D')
        content = input_file.read()
    else:
        content = input_file.read()
        input_file.close()

    return content


if __name__ == '__main__':
    main()



"""Machine learning model"""

from copy import deepcopy
import logging
from operator import itemgetter
from pathlib import Path
import shutil
from tempfile import TemporaryDirectory
from typing import List, Tuple, Dict, Any, Callable

import tensorflow as tf
from tensorflow.estimator import ModeKeys, Estimator
from tensorflow.python.training.tracking.tracking import AutoTrackable


LOGGER = logging.getLogger(__name__)

DATASET = {
    ModeKeys.TRAIN: 'train',
    ModeKeys.EVAL: 'valid',
    ModeKeys.PREDICT: 'test',
}


class HyperParameter:
    """Model hyper parameters"""
    BATCH_SIZE = 100
    NB_TOKENS = 1000
    VOCABULARY_SIZE = 20000
    EMBEDDING_SIZE = int(VOCABULARY_SIZE**0.5)
    DNN_HIDDEN_UNITS = [512, 32]
    DNN_DROPOUT = 0.5


class Training:
    """Model training parameters"""
    SHUFFLE_BUFFER = HyperParameter.BATCH_SIZE * 10
    CHECKPOINT_STEPS = 1000
    LONG_TRAINING_STEPS = 10 * CHECKPOINT_STEPS
    SHORT_DELAY = 60
    LONG_DELAY = 5 * SHORT_DELAY


def load(saved_model_dir: str) -> AutoTrackable:
    """Load a Tensorflow saved model"""
    return tf.saved_model.load(saved_model_dir)


def build(model_dir: str, labels: List[str]) -> Estimator:
    """Build a Tensorflow text classifier """
    config = tf.estimator.RunConfig(
        model_dir=model_dir,
        save_checkpoints_steps=Training.CHECKPOINT_STEPS,
    )
    categorical_column = tf.feature_column.categorical_column_with_hash_bucket(
        key='content',
        hash_bucket_size=HyperParameter.VOCABULARY_SIZE,
    )
    dense_column = tf.feature_column.embedding_column(
        categorical_column=categorical_column,
        dimension=HyperParameter.EMBEDDING_SIZE,
    )

    return tf.estimator.DNNLinearCombinedClassifier(
        linear_feature_columns=[categorical_column],
        dnn_feature_columns=[dense_column],
        dnn_hidden_units=HyperParameter.DNN_HIDDEN_UNITS,
        dnn_dropout=HyperParameter.DNN_DROPOUT,
        label_vocabulary=labels,
        n_classes=len(labels),
        config=config,
    )


def train(estimator: Estimator, data_root_dir: str, max_steps: int) -> Any:
    """Train a Tensorflow estimator"""

    train_spec = tf.estimator.TrainSpec(
        input_fn=_build_input_fn(data_root_dir, ModeKeys.TRAIN),
        max_steps=max_steps,
    )

    if max_steps > Training.LONG_TRAINING_STEPS:
        throttle_secs = Training.LONG_DELAY
    else:
        throttle_secs = Training.SHORT_DELAY

    eval_spec = tf.estimator.EvalSpec(
        input_fn=_build_input_fn(data_root_dir, ModeKeys.EVAL),
        start_delay_secs=Training.SHORT_DELAY,
        throttle_secs=throttle_secs,
    )

    LOGGER.debug('Train the model')
    results = tf.estimator.train_and_evaluate(estimator, train_spec, eval_spec)
    training_metrics = results[0]
    return training_metrics


def save(estimator: Estimator, saved_model_dir: str) -> None:
    """Save a Tensorflow estimator"""
    with TemporaryDirectory() as temporary_model_base_dir:
        export_dir = estimator.export_saved_model(
            temporary_model_base_dir, _serving_input_receiver_fn
        )

        Path(saved_model_dir).mkdir(exist_ok=True)
        export_path = Path(export_dir.decode()).absolute()
        for path in export_path.glob('*'):
            shutil.move(str(path), saved_model_dir)


def test(
    saved_model: AutoTrackable,
    data_root_dir: str,
    mapping: Dict[str, str],
) -> Dict[str, Dict[str, int]]:
    """Test a Tensorflow saved model"""
    values = {language: 0 for language in mapping.values()}
    matches = {language: deepcopy(values) for language in values}

    input_function = _build_input_fn(data_root_dir, ModeKeys.PREDICT)
    for test_item in input_function():
        content = test_item[0]
        label = test_item[1].numpy()[0].decode()

        result = saved_model.signatures['predict'](content)
        predicted = result['classes'].numpy()[0][0].decode()

        label_language = mapping[label]
        predicted_language = mapping[predicted]
        matches[label_language][predicted_language] += 1

    return matches


def predict(
    saved_model: AutoTrackable,
    mapping: Dict[str, str],
    text: str
) -> List[Tuple[str, float]]:
    """Infer a Tensorflow saved model"""
    content_tensor = tf.constant([text])
    predicted = saved_model.signatures['serving_default'](content_tensor)

    numpy_floats = predicted['scores'][0].numpy()
    extensions = predicted['classes'][0].numpy()

    probability_values = (float(value) for value in numpy_floats)
    languages = (mapping[ext.decode()] for ext in extensions)

    unsorted_scores = zip(languages, probability_values)
    scores = sorted(unsorted_scores, key=itemgetter(1), reverse=True)
    return scores


def _build_input_fn(
    data_root_dir: str,
    mode: ModeKeys,
) -> Callable[[], tf.data.Dataset]:
    """Generate an input fonction for a Tensorflow model"""
    pattern = str(Path(data_root_dir).joinpath(DATASET[mode], '*'))

    def input_function() -> tf.data.Dataset:
        dataset = tf.data.Dataset
        dataset = dataset.list_files(pattern, shuffle=True).map(_read_file)

        if mode == ModeKeys.PREDICT:
            return dataset.batch(1)

        if mode == ModeKeys.TRAIN:
            dataset = dataset.shuffle(Training.SHUFFLE_BUFFER).repeat()

        return dataset.map(_preprocess).batch(HyperParameter.BATCH_SIZE)

    return input_function


def _serving_input_receiver_fn() -> tf.estimator.export.ServingInputReceiver:
    """Function to serve model for predictions."""

    content = tf.compat.v1.placeholder(tf.string, [None])
    receiver_tensors = {'content': content}
    features = {'content': tf.map_fn(_preprocess_text, content)}

    return tf.estimator.export.ServingInputReceiver(
        receiver_tensors=receiver_tensors,
        features=features,
    )


def _read_file(filename: str) -> Tuple[tf.Tensor, tf.Tensor]:
    """Read a source file, return the content and the extension"""
    data = tf.io.read_file(filename)
    label = tf.strings.split([filename], '.').values[-1]
    return data, label


def _preprocess(
    data: tf.Tensor,
    label: tf.Tensor,
) -> Tuple[Dict[str, tf.Tensor], tf.Tensor]:
    """Process input data as part of a workflow"""
    data = _preprocess_text(data)
    return {'content': data}, label


def _preprocess_text(data: tf.Tensor) -> tf.Tensor:
    """Feature engineering"""
    padding = tf.constant(['']*HyperParameter.NB_TOKENS)
    data = tf.strings.split([data]).values
    data = tf.concat((data, padding), axis=0)
    data = data[:HyperParameter.NB_TOKENS]
    return data



#!/usr/bin/env python3
"""
Setup Guesslang

* Install with pip (recommended):
    pip install .

* Install with setuptools:
    pip install -r requirements.txt
    python setup.py install

* Run tests:
    python setup.py pytest

"""

import ast
from pathlib import Path
import re

from setuptools import setup, find_packages


def version(base_module: str) -> str:
    version_pattern = r'__version__\s+=\s+(.*)'
    init_path = Path(Path(__file__).parent, base_module, '__init__.py')
    found = re.search(version_pattern, init_path.read_text())
    if not found:
        raise RuntimeError(f'{base_module} version not found')

    repr_value = found.group(1)
    return str(ast.literal_eval(repr_value))


def long_description(filename: str, end_tag: str, doc_url: str) -> str:
    lines = []
    for line in Path(filename).read_text().splitlines():
        if end_tag in line:
            break

        lines.append(line)

    lines.append('Full documentation at {}'.format(doc_url))
    return '\n'.join(lines)


setup(
    # Package info
    name='guesslang',
    author='Y. SOMDA',
    url='https://github.com/yoeo/guesslang',
    description='Detect the programming language of a source code',
    long_description=long_description(
        'docs/contents.rst',
        'end-description',
        'https://guesslang.readthedocs.io/en/latest/',
    ),
    license='MIT',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    # Install setup
    version=version('guesslang'),
    platforms='any',
    packages=find_packages(exclude=['tests', 'tools']),
    install_requires=Path('requirements.txt').read_text(),
    zip_safe=False,
    include_package_data=True,
    entry_points={
        'console_scripts': ['guesslang = guesslang.__main__:main']
    },
    # Test setup
    tests_require=Path('requirements-dev.txt').read_text(),
    setup_requires=['pytest-runner']
)
